#!/usr/local/bin/perl
# dtd.pl -- perl parser for SGML DTD's
# $Id: dtd.pl,v 1.1 1995/07/14 15:57:33 connolly Exp $
#


############
#
# MAIN
#
############

				# command line opts
while(($opt = $ARGV[0]) =~ /^-/){
    shift;

    if($opt eq '-i'){		# ala sgmls -i foo
	local($ent);
	$ent = shift;
	$Parameter{$ent} = 'INCLUDE';
    }
    elsif($opt eq '-d'){
	$Debug = shift;
    }
    else{
	warn "flag `$opt' ignored";
    }
}

while($ARGV[0] =~ /(\w+)=(.*)/){
    eval("\$$1=\$2");
    shift;
}

&parseDTD;			# Compute %Model

@Elements = sort keys %Model;

foreach(@Elements){		# Compute %Content
    local($model) = $Model{$_};

    if($model =~ /\#PCDATA/ || $model eq 'ANY'){
	$Content{$_} = 'MIXED';
    }elsif($model =~ /EMPTY|CDATA|RCDATA/){
	$Content{$_} = $model;
    }else{
	$Content{$_} = 'ELEMENT';
    }
}

#&summary;
&giarray;
#&tagList; # for MidasWWW app-defaults file

sub giarray{

    foreach(@Elements){
	local($content) = $Content{$_};
	next if $content eq 'EMPTY';
	print <<EOF;
	{"$_", (char*)SGML_$content},
EOF
    }
}


sub tagList{

    foreach(@Elements){		# write midasWWW tagList
	next if $Content{$_} =~ /^(EMPTY|RCDATA|CDATA)$/;
	
	print "*", $Doctype, "*", $_, ".tagList:";
	
	local(%taglist);
	
	foreach(split(/\s+/, $Model{$_})){
	    next if /#/;
	    if($Content{$_} eq 'MIXED'){
		$taglist{"SGMLCompoundText:"} .= " ". $_;
	    }elsif($Content{$_} eq 'ELEMENT'){
		$taglist{"SGMLText:"} .= " ". $_;
	    }elsif($Content{$_} eq 'EMPTY'){
		$taglist{"SGMLMarkerText:"} .= " ". $_;
	    }else{
		# CDATA or RCDATA
		$taglist{"SGMLFormattedText:"} .= " ". $_;
	    }
	}
	
	foreach(sort keys %taglist){
	    print "\\\n\t", $_, $taglist{$_};
	}
	
	print "\n\n";
    }
}

    
sub summary{

    foreach(sort @Elements){		# write human-readable summary
	print "\n*** ", $_, ":\n";

	local($c) = $Content{$_};

	if($c eq 'MIXED'){
	    print "MIXED content: data characters, elements, comments, other markup...\n";
	    print "Elements allowed inside: ", $Model{$_}, "\n";
	}
	elsif($c eq 'ELEMENT'){
	    print "ELEMENT content: no data. only elements and other markup\n";
	    print "Elements allowed inside: ", $Model{$_}, "\n";
	}
	elsif($c eq 'RCDATA'){
	    print "RCDATA content: data characters and entity references only\n";
	}
	elsif($c eq 'CDATA'){
	    print "CDATA content: data characters only\n";
	}
	elsif($c eq 'EMPTY'){
	    print "EMPTY: no content or end tag\n";
	}

	print "\n";
    }
}

    
#####
#
# parsing routines
#
#####

sub parseDTD{
    while(($_ = &lex) gt ''){
	&debug(5, "token: ", $_);
	
	last if !defined($_) || $_ eq ']>';
	
	if(/<!SGML/){
	    while(defined($_)){
		$_ = &lex;
		last if $_ eq '>'; #@@ skip SGML declaration
	    }
	}
	elsif(/<!DOCTYPE/){
	    $Doctype = &lex;
	    last unless &lex eq '[';
	}
	elsif(/<!ENTITY/){
	    local($name) = &lex;
	    local($val) = &lex;
	    
	    if($name eq '%'){
		$name = $val;
		$val = &lex;
		
		if(!($val =~ s/^\"//)){
		    warn "entity value not supported: ", $val;
		    next;
		}
		$val =~ s/\"$//;

		# only the first definition of an entity counts
		if($Parameter{$name}){
		    &debug(3, " definition of $name as `$val' ignored");
		}else{
		    &debug(2, "entity $name = `$val'");
		    $Parameter{$name} = $val;
		}
	    }else{
		$General{$name} = $val;
	    }

	    $val = &lex;
	    warn "expected >, got: `", $val, "'" unless $val eq '>';
	}
	
	elsif(/<!ELEMENT/){
	    local(@names) = &namegroup;
	    local($smin) = &lex;
	    local($emin) = &lex;
	    local(@model) = &namegroup;
	    
	    grep(tr/a-z/A-Z/, @model) if $Fold;
	    
	    &debug(2, "names: ", join(' ', @names), "\nmodel: "
		   ,join(' ', @model));
	    
	    foreach(@names){
		tr/a-z/A-Z/ if $Fold;
		$Model{$_} = join(' ', sort(@model));
	    }
	    
	    $_ = &lex; # eat MDC (i.e. >)
	    $_ = &lex if $_ eq '*' || $_ eq '?' || $_ eq '+'; # occurence
	}
    }
}

sub namegroup{
    local(@ret);
    local($name) = &lex;
    local($depth) = ($name eq '(');

    push(@ret, $name) unless $depth;

    while($depth > 0){
	$name = &lex;
	last if name le '';
	$depth++ if $name eq '(';
	$depth-- if $name eq ')';
	push(@ret, $name) if $name =~ /[a-zA-Z][\w\.-]*/;
    }
    return @ret;
}


sub lex{
    local($ret);
    
    do{
	do{
	    &lexprep || return undef;
	    
	    # whitespace
	    $E[0] =~ s/^\s+//;
	}while($E[0] le '');

	&debug(6, "analyzing: `$E[0]'");

	# parameter entity reference
	if($E[0] =~ s/^%([a-zA-Z][\w\.-]*);?//){
	    warn $1, " undefined" unless defined($Parameter{$1});
	    &debug(4, "expanding entity $1 to `$Parameter{$1}'");
	    unshift(@E, $Parameter{$1});
	}
	
	# Marked section
	elsif($E[0] =~ s/^<!\[\s*//){
	    local($tok, $status, $s, @ms);

	    $status = &lex;

	    $s = $status =~ /^INCLUDE$/i ? 1 : 0;

	    &debug(1, "Marked section status: $status ($s)");

	    warn "expected [, got: `$tok'" unless ($tok = &lex) eq '[';

	    while(! ($E[0] =~ m/\]\]>/)){
		push(@ms, $E[0]) if $s;
		undef($E[0]);
		&lexprep || last;
	    }
	    $E[0] =~ m/\]\]>/;
	    $E[0] = $';
	    push(@ms, $`) if $s;
	    unshift(@E, @ms) if $s;
	    &lexprep || return undef;
	}

	# comment declaration
	elsif($E[0] =~ s/^<!--//){
	    while(! ($E[0] =~ m/--/)){
		undef($E[0]);
		&lexprep || last;
	    }
	    $E[0] =~ m/--/;
	    $E[0] = $';
	    &lexprep || return undef;
	    $E[0] =~ s/^\s*>//;
	}
	
	# comment
	elsif($E[0] =~ s/^--//){
	    while(! ($E[0] =~ m/--/)){
		undef($E[0]);
		&lexprep || last;
	    }
	    $E[0] =~ m/--/;
	    $E[0] = $';
	}
	
	# markup declaration
	elsif($E[0] =~ s/^(<![a-zA-Z][\w\.-]*)//){
	    $ret = $1;
	    $ret =~ tr/a-z/A-Z/ if $Fold;
	}
	# reserved name
	elsif($E[0] =~ s/^(\#[a-zA-Z][\w\.-]*)//){
	    $ret = $1;
	    $ret =~ tr/a-z/A-Z/ if $Fold;
	}

	# delimiter
	elsif($E[0] =~
	      s/^(\||\&|\?|\+|\,|\-|\[|\(|\)|\*|\>|\%|(\]\]>)|(\]>))//){
	    $ret = $&;
	}

	# name
	elsif($E[0] =~ s/^[a-zA-Z][\w\.-]*//){
	    $ret = $&;
	}

	# number
	elsif($E[0] =~ s/^\d+//){
	    $ret = $&;
	}

	# string
	elsif($E[0] =~ s/^\"//){
	    $ret = '"';
	    while(! ($E[0] =~ s/^([^\"]*)\"//)){
		$ret .= $E[0];
		undef($E[0]);
		&lexprep || last;
		&debug(9, "looking for end of string: `$E[0]'");
	    }
	    $ret .= $1 . '"';
	    &debug(9, "left over after string: `$E[0]'");
	}
	else{
	    warn "unrecognized token: `", $E[0], "'";
	    undef($E[0]);
	}
    }while($ret le '');

    &debug(9, "token: `$ret' next: `$E[0]'");
    return $ret;
}

sub lexprep{
    while($E[0] le ''){
	if($#E > 0){
	    shift @E;
	}else{
	    local($_);
	    # all out of stuff.. read in the next line
	    $_ = <>;
	    s/\n/ /g; # because we don't want ^ in s/^xxx/ to match newlines
	    $E[0] = $_;
	    &debug(7, $., ":", $E[0]);
	    return 0 if eof;
	}
    }
    return 1;
}

sub debug{
    local($level, @strings) = @_;

    print STDERR @strings, "\n" if $Debug >= $level;
}
